from torch.utils.data import Dataset

class CellDataset(Dataset):
    def __init__(self, y, x, s):
        self.y = y
        self.x = x
        self.s = s

    def __len__(self):
        return self.y.shape[0]

    def __getitem__(self, idx):
        item = {"Y": self.y[idx, :], "X": self.x[idx, :], "s": self.s[idx]}
        return item

if __name__ == '__main__':
    from torch.utils.data import DataLoader
    from scipy.io import mmread
    import numpy as np
    import pandas as pd

    Y_=mmread('../datasets/koh_extract/koh.data.counts.mm').toarray().astype(float).transpose()
    s_=pd.read_csv('../datasets/analysis/koh.size_factor_cluster.tsv',sep='\t',header=None)[0].values.astype(float)#.reshape(-1,1)
    X_=np.array([np.ones(Y_.shape[0])]).transpose()

    cell_dataset = CellDataset(Y_, X_, s_)
    cell_dataloader = DataLoader(dataset=cell_dataset, shuffle=False, batch_size=Y_.shape[0], num_workers=0)



import datetime
import multiprocessing

import anndata as ad
import numpy as np
import pandas as pd
from typing import Union, List, Tuple

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from tqdm import tqdm
# from tqdm import tqdm_notebook as tqdm

from MarcoPolo.regression.models import MarcoPoloModel
from MarcoPolo.regression.datasets import CellDataset

torch.set_default_dtype(torch.float64)


def fit_one_gene(model: nn.Module, optimizer: optim.Adamax, cell_dataloader: DataLoader, device: str,
                 EM_ITER_MAX: float, M_ITER_MAX: float, LL_diff_tolerance: float, Q_diff_tolerance: float,
                 verbose: bool = True):
    """
    Run EM trick algorithm.
    Args:
        model: Model to be trained
        optimizer: Optimizer to be used
        cell_dataloader: DataLoader for training
        device: device to use. If you want to use GPU set to 'cuda:0'. If you want to use CPU set to 'cpu'
        EM_ITER_MAX: maximum number of iterations of E-step of the EM trick algorithm
        M_ITER_MAX: maximum number of iterations of M-step of the EM trick algorithm
        LL_diff_tolerance: tolerance for the difference of log likelihood between two iterations of the EM trick algorithm
        Q_diff_tolerance: tolerance for the difference of Q between two iterations of the EM trick algorithm
        verbose:

    Returns:
        gamma_new: gamma after EM trick algorithm
        LL_new: log likelihood after EM trick algorithm
        Q_new: Q after EM trick algorithm
        em_idx_max: number of iterations of the EM trick algorithm
        m_idx_max: number of iterations of the M-step of the EM trick algorithm
    """
    if verbose:
        print('Start time:', datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))

    with torch.no_grad():
        for batch_idx, batch in enumerate(cell_dataloader):
            batch_Y = batch['Y'].to(device)
            batch_X = batch['X'].to(device)
            batch_s = batch['s'].to(device)
        LL_old = model(batch_Y, batch_X, batch_s)
        Q_old = LL_old
    if verbose:
        print(LL_old)

    em_idx_max = 0
    m_idx_max = 0

    for em_idx in range(EM_ITER_MAX):  #
        LL_new = torch.zeros_like(LL_old)
        for batch_idx, batch in enumerate(cell_dataloader):
            # Usually batch size is the size of samples and only one batch is used. But if the dataset is too large, it is better to use multiple batches.
            batch_Y = batch['Y'].to(device)
            batch_X = batch['X'].to(device)
            batch_s = batch['s'].to(device)

            #############
            # M-step
            #############
            for m_idx in range(M_ITER_MAX):
                optimizer.zero_grad()
                Q_new = -model(batch_Y, batch_X, batch_s)
                Q_new.backward()
                optimizer.step()

                # Constraint
                model.delta_log.data = model.delta_log.data.clamp(min=model.delta_log_min)
                # model.NB_basis_a.data=model.NB_basis_a.data.clamp(min=0)

                if m_idx % 20 == 0:
                    Q_diff = (Q_old - Q_new) / torch.abs(Q_old)
                    Q_old = Q_new
                    if verbose:
                        print('M: {}, Q: {} Q_diff: {}'.format(m_idx, Q_new, Q_diff))
                    if m_idx > 0 and torch.abs(Q_diff) < (Q_diff_tolerance):
                        if verbose:
                            print('M break')
                        break
            m_idx_max = max(m_idx_max, m_idx)

            #############
            # Look at LL
            #############
            with torch.no_grad():
                LL_temp = -Q_new
                LL_new += LL_temp

        LL_diff = (LL_new - LL_old) / torch.abs(LL_old)
        LL_old = LL_new

        if verbose:
            print('EM: {}, LL: {} LL_diff: {}'.format(em_idx, LL_new, LL_diff))
        if LL_diff < LL_diff_tolerance:
            if verbose:
                print('EM break')
            break
    em_idx_max = max(em_idx_max, em_idx)

    if verbose:
        print('End time:', datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))

    with torch.no_grad():
        gamma_new = model(batch_Y, batch_X, batch_s, to_return='gamma')

    return gamma_new, Q_new, LL_new, em_idx_max, m_idx_max


def fit_multiple_genes(Y: np.array, X: np.array, s: np.array,
                       num_cluster_list: List, learning_rate: float, fit_one_gene_parameters: dict, device: str,
                       start_gene_idx: Union[None, int] = None, end_gene_idx: Union[None, int] = None,
                       verbose: bool = False)-> dict:
    """
    Fit multiple genes.

    Args:
        Y (np.array): matrix of gene expression (cell, gene)
        X (np.array): matrix of cell covariates (cell, feature)
        s (np.array): vector of cell sizes (cell, )
        num_cluster_list (list): list of number of clusters to test
        learning_rate (float): learning rate of the optimizer
        fit_one_gene_parameters (dict): parameters for fit_one_gene
        start_gene_idx (Union[None, int]): start gene index. required for multithreading. If None, start from the first gene.
        end_gene_idx (Union[None, int]): end gene index. required for multithreading.  If None, end at the last gene.
        device (str): device to use. If you want to use GPU set to 'cuda:0'. If you want to use CPU set to 'cpu'
        verbose (bool): if True, print out the progress

    Returns:
        regression_result: a dictionary containing the regression results with the following fields: `gamma_list_cluster`, `delta_log_cluster`, `beta_cluster`, `result_cluster`.

    """

    if start_gene_idx is not None or end_gene_idx is not None:
        Y_select = Y[:, start_gene_idx:end_gene_idx]
    else:
        Y_select=Y

    device = torch.device(device)

    gamma_cluster = {}

    Q_cluster = {}
    LL_cluster = {}
    em_idx_max_cluster = {}
    m_idx_max_cluster = {}

    delta_log_cluster = {}
    beta_cluster = {}

    for idx, num_cluster in enumerate(num_cluster_list):
        gamma_list = []

        Q_list = []
        LL_list = []
        em_idx_max_list = []
        m_idx_max_list = []

        delta_log_list = []
        beta_list = []

        if len(multiprocessing.current_process()._identity) == 0 or multiprocessing.current_process()._identity[0] == 1:
            print(f'({idx+1}) Fitting with {num_cluster} cluster(s)')
            pbar = tqdm(np.arange(Y_select.shape[1]), desc='Progress')
        else:
            pbar = np.arange(Y_select.shape[1])

        for iter_idx, exp_data_idx in enumerate(pbar):
            cell_dataset = CellDataset(Y_select[:, iter_idx:iter_idx + 1], X, s)

            cell_dataloader = DataLoader(dataset=cell_dataset, shuffle=False, batch_size=Y_select.shape[0],
                                         num_workers=0)

            if iter_idx == 0:
                model = MarcoPoloModel(Y=Y_select[:, iter_idx:iter_idx + 1], rho=np.ones((num_cluster, 1)),
                                       X_col=X.shape[1],
                                       delta_min=0).to(device)
            else:
                model.init_parameter_delta_min(0)
                model.init_paramter_Y(Y_select[:, iter_idx:iter_idx + 1])
            optimizer = optim.Adamax(model.parameters(), lr=learning_rate)  # ,betas=(0.92, 0.999))
            gamma, Q, LL, em_idx_max, m_idx_max = fit_one_gene(model=model, optimizer=optimizer,
                                                               cell_dataloader=cell_dataloader, device=device,
                                                               **fit_one_gene_parameters, verbose=verbose)

            gamma_list.append(gamma.cpu().numpy())

            Q_list.append(Q.detach().cpu().numpy())
            LL_list.append(LL.detach().cpu().numpy())
            em_idx_max_list.append(em_idx_max)
            m_idx_max_list.append(m_idx_max)

            delta_log_list.append(model.delta_log.detach().cpu().numpy())
            beta_list.append(model.beta.detach().cpu().numpy())

        gamma_cluster[num_cluster] = gamma_list

        delta_log_cluster[num_cluster] = delta_log_list

        beta_cluster[num_cluster] = beta_list

        Q_cluster[num_cluster] = Q_list
        LL_cluster[num_cluster] = LL_list
        em_idx_max_cluster[num_cluster] = em_idx_max_list
        m_idx_max_cluster[num_cluster] = m_idx_max_list

    result_cluster = {num_cluster: pd.DataFrame([Q_cluster[num_cluster],
                                                 LL_cluster[num_cluster],
                                                 em_idx_max_cluster[num_cluster],
                                                 m_idx_max_cluster[num_cluster]],
                                                index=['Q', 'LL', 'em_idx_max', 'm_idx_max']).T
                      for num_cluster in num_cluster_list}

    regression_result = {"gamma_list_cluster": gamma_cluster,
                         "delta_log_cluster": delta_log_cluster,
                         "beta_cluster": beta_cluster,
                         "result_cluster": result_cluster, }
    return regression_result


def run_regression(adata: ad.AnnData, size_factor_key: Union[str, None], covariates=None,
                   num_cluster_list=[1, 2], learning_rate=0.1,
                   EM_ITER_MAX=20, M_ITER_MAX=10000, LL_diff_tolerance=1e-4, Q_diff_tolerance=1e-4,
                   device: str='cuda:0', num_threads=1, verbose=False)->dict:
    """
    Run regression.

    Args:
        adata: anndata.AnnData containing scRNA-seq data. `.X` should be a matrix containing raw count data of shape (# cells, # genes).
        size_factor_key: key of the size factor stored in `adata.obs`. If not set, you can calculate size factor using `scanpy.pp.normalize_total` as follows. `norm_factor = sc.pp.normalize_total(adata, exclude_highly_expressed=True, max_fraction= 0.2, inplace=False)["norm_factor"]; adata.obs["size_factor"] = norm_factor/norm_factor.mean()` If None, no size factor is used.
        covariates: a covariate matrix of shape (# cells, # covariates). Default: None.
        num_cluster_list: a list of numbers of clusters to test. Default: [1, 2].
        learning_rate: learning rate of the Adamax optimizer. Default: 0.1.
        EM_ITER_MAX: maximum number of iterations of E-step of the EM trick algorithm. Default: 20.
        M_ITER_MAX: maximum number of iterations of M-step of the EM trick algorithm. Default: 10000.
        LL_diff_tolerance: tolerance of the difference of log-likelihood between two iterations of the EM trick algorithm. Default: 1e-4.
        Q_diff_tolerance: tolerance of the difference of Q between two iterations of the EM trick algorithm. Default: 1e-4.
        device: device to use. If you want to use GPU set to 'cuda:0'. If you want to use CPU set to 'cpu' Default: 'cuda:0'.
        verbose: if True, print the progress of the EM trick algorithm. Default: False.
        num_threads: number of threads to use. For each gene, MarcoPolo fits Poisson model to a matrix of (1, # cells). As the matrix is too small for us to fully utilize the power of GPU, it is good to use multiple threads at once. The best option depends on the number of cells and the GPU memory size. For 500 cells and 11GB, using 4 threads worked well. Default: 1.

    Returns:
        regression_result: a dictionary containing the regression results with the following fields: `gamma_list_cluster`, `delta_log_cluster`, `beta_cluster`, `result_cluster`.

    """
    if num_threads > 1 and device.startswith('cuda'):
        print(
            f"<INFO> Currently, you are using {num_threads} threads for regression. If you encounter any memory issues, try to set `num_threads` to 1.")

    expression_matrix = adata.X  # (cell, gene)
    num_cells = expression_matrix.shape[0]
    num_genes = expression_matrix.shape[1]

    if not type(expression_matrix) == np.ndarray:
        expression_matrix = expression_matrix.toarray().astype(float)
    else:
        expression_matrix = expression_matrix.astype(float)

    if size_factor_key is None:
        cell_size_factor = np.ones(expression_matrix.shape[0]).astype(float)
    else:
        cell_size_factor = adata.obs[size_factor_key].values.astype(float)

    if covariates is None:
        covariate_matrix = np.ones((expression_matrix.shape[0], 1)).astype(float)
    else:
        covariate_matrix = covariates.astype(float)

    fit_one_gene_parameters = {"EM_ITER_MAX": EM_ITER_MAX, "M_ITER_MAX": M_ITER_MAX,
                               "LL_diff_tolerance": LL_diff_tolerance, "Q_diff_tolerance": Q_diff_tolerance}


    print(f'The numbers of clusters to test: {num_cluster_list}')
    print(f'Y: {expression_matrix.shape} X: {covariate_matrix.shape} s: {cell_size_factor.shape}')

    if num_threads != 1:
        multiprocessing.set_start_method('spawn', force=True)
        pool = multiprocessing.Pool(processes=num_threads)

        gene_per_thread = expression_matrix.shape[1] // num_threads
        gene_thread_split = [(gene_per_thread * i, gene_per_thread * (i + 1)) for i in range(num_threads - 1)] + [
            (gene_per_thread * (num_threads - 1), expression_matrix.shape[1])]
        #gene_thread_split=gene_thread_split[::-1]

        #multiprocessing.freeze_support()

        fit_result_thread = pool.starmap(fit_multiple_genes, [(expression_matrix[:, start_gene_idx: end_gene_idx],
                                                               covariate_matrix[:],
                                                               cell_size_factor[:],
                                                               num_cluster_list,
                                                               learning_rate,
                                                               fit_one_gene_parameters,
                                                               device,
                                                               None,
                                                               None,
                                                               verbose) for start_gene_idx, end_gene_idx in
                                                              gene_thread_split])

        pool.close()

        regression_result = {}

        for fit_result_thread in fit_result_thread[:]:
            for category, value_cluster in fit_result_thread.items():
                for num_cluster, value in value_cluster.items():
                    if isinstance(value, list):
                        regression_result[category][num_cluster] = regression_result.setdefault(category, {}).get(
                            num_cluster, []) + value
                    elif isinstance(value, pd.DataFrame):
                        regression_result[category][num_cluster] = regression_result.setdefault(category, {}).get(
                            num_cluster, []) + [value.reset_index()]
                    else:
                        raise ValueError("Unknown type of value: {}".format(type(value)))

        for category in regression_result.keys():
            for num_cluster in regression_result[category].keys():
                if isinstance(regression_result[category][num_cluster][0], pd.DataFrame):
                    regression_result[category][num_cluster] = pd.concat(
                        regression_result[category][num_cluster]).reset_index()
                assert len(regression_result[category][num_cluster])==num_genes, RuntimeError("Length of result is not equal to number of genes.")

    else:
        regression_result = fit_multiple_genes(Y=expression_matrix[:, :],
                                               X=covariate_matrix[:],
                                               s=cell_size_factor[:],
                                               num_cluster_list=num_cluster_list,
                                               learning_rate=learning_rate,
                                               fit_one_gene_parameters=fit_one_gene_parameters,
                                               device=device,
                                               verbose=verbose)

    return regression_result


if __name__ == '__main__':
    data_path = "/homes/gws/chanwkim/MarcoPolo/notebooks/example/hESC.h5ad"
    adata = ad.read(data_path)
    run_regression(adata=adata[:, :10], size_factor_key="size_factor", num_threads=3, device="cuda:2")

    # from scipy.io import mmread
    # import numpy as np
    # import pandas as pd
    # Y_=mmread('../datasets/koh_extract/koh.data.counts.mm').toarray().astype(float).transpose()
    # s_=pd.read_csv('../datasets/analysis/koh.size_factor_cluster.tsv',sep='\t',header=None)[0].values.astype(float)#.reshape(-1,1)
    # X_=np.array([np.ones(Y_.shape[0])]).transpose()

    device = torch.device('cuda:2')

    Y_ = np.ones((446, 4898))
    X_ = np.ones((446, 1))
    s_ = np.ones((446))

    fit_multiple_genes(Y_select=Y_, X=X_, s=s_, num_cluster_list=[1, 2, 3], LR=0.1, EM_ITER_MAX=20, M_ITER_MAX=10000,
                       LL_diff_tolerance=1e-4, Q_diff_tolerance=1e-4, device=device, verbose=True)



import multiprocessing
import warnings

import numpy as np
import pandas as pd


def gamma_argmax_list_to_oncell_size_list_list(gamma_argmax_list: np.ndarray)->np.ndarray:
    """

    Args:
        gamma_argmax_list: List of gamma_argmax.

    Returns:
        np.ndarray: List of oncell size.

    """
    oncellsize_list=np.sum(gamma_argmax_list==0,axis=1)
    # print(oncellsize_list)
    oncellsize_list_list=[np.clip(oncellsize_list, a_min=0, a_max=oncellsize) for oncellsize in oncellsize_list]

    return np.array(oncellsize_list_list)


def gamma_argmax_list_to_intersection(gamma_argmax_list, idx):
    """

    Args:
        gamma_argmax_list: List of gamma_argmax.
        idx: Index of gamma_argmax_list.

    Returns:
        np.ndarray: Intersection.

    """
    intersection = np.sum((gamma_argmax_list[idx] == gamma_argmax_list) & (gamma_argmax_list[idx] == 0), axis=1)
    return intersection

def gamma_argmax_list_to_intersection_list(gamma_argmax_list: np.ndarray)->np.ndarray:
    """

    Args:
        gamma_argmax_list:  List of gamma_argmax.

    Returns:
        np.ndarray: List of intersection.

    """

    pool=multiprocessing.Pool(processes=16)

    intersection_list=pool.starmap(gamma_argmax_list_to_intersection,[(gamma_argmax_list,i) for i in np.arange(gamma_argmax_list.shape[0])])

    pool.close()
    pool.join()

    return np.array(intersection_list)

def gamma_expression_to_gamma_argmax(gamma: np.ndarray, expression: np.ndarray = None) -> np.ndarray:
    """

    Args:
        gamma: A gamma matrix.
        expression: If expression is not None, it is used to calculate the which group has higher expression mean.

    Returns:
        np.ndarray: gamma_argmax.

    """
    gamma_argmax = np.argmax(gamma, axis=1)
    gamma_argmax_counts = list(np.unique(gamma_argmax, return_counts=True))
    if expression is None:
        key_newkey = pd.DataFrame(gamma_argmax_counts, index=['idx', 'counts']).T.set_index('idx').sort_values(
            by='counts', ascending=True).index.tolist()
    else:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            gamma_argmax_counts_lfc = gamma_argmax_counts + [(list(
                map(lambda x: np.mean(expression[gamma_argmax == x], axis=0) - np.mean(
                    expression[gamma_argmax != x], axis=0), gamma_argmax_counts[0])))]
        key_newkey = pd.DataFrame(gamma_argmax_counts_lfc, index=['idx', 'counts', 'lfc']).T.astype(
            {'idx': int, 'counts': int}).set_index('idx').sort_values(by='lfc', ascending=False).index.tolist()
    gamma_argmax = [key_newkey.index(argmax) for argmax in gamma_argmax]

    return gamma_argmax

def gamma_list_expression_matrix_to_gamma_argmax_list(gamma_list: np.ndarray, expression_matrix: np.ndarray=None)->np.ndarray:
    """

    Args:
        gamma_list: List of gamma matrices.
        expression_matrix: If expression_matrix is not None, it is used to calculate the which group has higher expression mean.

    Returns:
        List of gamma_argmax.

    """
    pool=multiprocessing.Pool(processes=16)

    if expression_matrix is None:
        gamma_argmax_list=pool.map(gamma_expression_to_gamma_argmax, gamma_list)
    else:
        gamma_argmax_list=pool.starmap(gamma_expression_to_gamma_argmax, [(gamma_list[i], expression_matrix[:, i]) for i in range(len(gamma_list))])

    pool.close()
    pool.join()

    return np.array(gamma_argmax_list)


import numpy as np
import pandas as pd
import anndata as ad

import warnings

from sklearn import preprocessing
from sklearn.decomposition import PCA

import MarcoPolo.utils

pd.options.mode.chained_assignment = None

def find_markers(adata: ad.AnnData, regression_result: dict, mode: float=2, voting_thres: float=0.7, PCA_norm_thres:float=10, num_PC:int=2, log_fold_change_thres:float=0.6,
                 oncell_size_min_count:float=10, oncell_size_max_proportion:float=70)->pd.DataFrame:
    """
    find markers from the regression result
    Args:
        adata: anndata.AnnData containing scRNA-seq data. `.X` should be a matrix containing raw count data of shape (# cells, # genes).
        regression_result: dict containing regression results. Return value of `run_regression` function.
        mode: the number of groups to be used for marker selection. Default: 2.
        voting_thres: the threshold for voting. should be between 0 and 1. Default: 0.7.
        PCA_norm_thres: the threshold for PCA normalization. Default: 10.
        num_PC: the number of PCs to be used for marker selection. should be between 1 and 50 Default: 2.
        log_fold_change_thres: the threshold for log fold change. Default: 0.6.
        oncell_size_min_count: the minimum number of cells in on-cell group. Default: 10.
        oncell_size_max_proportion: the maximum proportion of cells in on-cell group. Default: 70.

    Returns:
        gene_scores: a pandas.DataFrame containing the following columns: 'MarcoPolo_rank', 'bimodality_score', 'voting_score', 'proximity_score', etc.

    """
    expression_matrix = adata.X.copy()
    num_cells=expression_matrix.shape[0]
    num_genes=expression_matrix.shape[1]

    ########################
    # Assign cells to on-cells and off-cells
    ########################
    print("Assign cells to on-cells and off-cells...")
    gamma_list = regression_result["gamma_list_cluster"][mode]
    gamma_argmax_list = MarcoPolo.utils.gamma_list_expression_matrix_to_gamma_argmax_list(gamma_list, expression_matrix)

    ########################
    # Calculate log fold change
    ########################
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        log_fold_change = np.log10(np.array([np.mean(expression_matrix[gamma_argmax_list[i] == 0, i]) for i in
                                 range(num_genes)]) / np.array(
            [np.mean(expression_matrix[gamma_argmax_list[i] != 0, i]) for i in range(num_genes)]))

    ########################
    # Calculate voting score
    ########################
    print("Calculating voting score...")
    oncell_size_list = np.sum(gamma_argmax_list == 0, axis=1)
    oncell_size_cliplist = MarcoPolo.utils.gamma_argmax_list_to_oncell_size_list_list(gamma_argmax_list)
    intersection_list = MarcoPolo.utils.gamma_argmax_list_to_intersection_list(gamma_argmax_list)
    #intersectioncount_prop=((intersection_list/oncell_size_cliplist))
    #intersectioncount_prop_top10=[np.arange(0,len(i))[i>=sorted(i)[-10]][:10] for i in intersectioncount_prop]
    intersectioncount_threshold = ((intersection_list / oncell_size_cliplist) > voting_thres)
    voting_score = np.sum(intersectioncount_threshold, axis=1)

    ########################
    # Calculate proximity score
    ########################
    print("Calculating proximity score...")
    expression_matrix_norm = np.log1p(10000 * expression_matrix / expression_matrix.sum(axis=1, keepdims=True))
    #expression_matrix_norm_scale = preprocessing.scale(expression_matrix_norm, axis=0, with_mean=True, with_std=True, copy=True)
    expression_matrix_norm_scale=(expression_matrix_norm-expression_matrix_norm.mean(axis=0, keepdims=True))/expression_matrix_norm.std(axis=0, keepdims=True)
    expression_matrix_norm_scale[expression_matrix_norm_scale > PCA_norm_thres] = PCA_norm_thres

    pca = PCA(n_components=50)
    pca.fit(expression_matrix_norm_scale)
    expression_matrix_norm_scale_pc = pca.transform(expression_matrix_norm_scale)

    proximity_score = np.array(
        [expression_matrix_norm_scale_pc[gamma_argmax_list[i] == 0, :num_PC].std(axis=0).mean() for i in
         range(num_genes)])

    ########################
    # Calculate bimodality score
    ########################
    print("Calculating bimodality score...")
    QQratio = regression_result["result_cluster"][1]['Q'] / regression_result["result_cluster"][mode]['Q']
    mean_all = np.array([np.mean(expression_matrix[:, i]) for i in range(num_genes)])

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        mean_on = np.array(
            [np.mean(expression_matrix[gamma_argmax_list[i] == 0, i]) for i in range(num_genes)])
    MS = mean_on - mean_all

    ########################
    # Final step of obtaining MarcoPolo score
    ########################
    print("Calculating MarcoPolo score...")
    gene_scores = pd.DataFrame([QQratio.values,
                                voting_score,
                                proximity_score,
                               log_fold_change,
                               MS,
                               oncell_size_list,],
                               index=['QQratio',
                                     'voting_score',
                                     'proximity_score',
                                     'log_fold_change',
                                     'MS',
                                     'oncell_size']).T

    gene_scores['QQratio_rank'] = \
        pd.Series(np.arange(num_genes), index=gene_scores['QQratio'].sort_values(ascending=False).index).loc[
            gene_scores.index]

    gene_scores['voting_score_rank'] = \
        pd.Series(np.arange(num_genes),
                  index=gene_scores['voting_score'].sort_values(ascending=False).index).loc[
            gene_scores.index]
    gene_scores['voting_score_rank'][gene_scores['voting_score'] == 0] = 499999
    gene_scores['voting_score_rank'][gene_scores['voting_score'] == 1] = 999999

    gene_scores['proximity_score_rank'] = \
        pd.Series(np.arange(num_genes), index=gene_scores['proximity_score'].sort_values(ascending=True).index).loc[
            gene_scores.index]

    gene_scores['log_fold_change_rank'] = \
        pd.Series(np.arange(num_genes), index=gene_scores['log_fold_change'].sort_values(ascending=False).index).loc[
            gene_scores.index]

    gene_scores['MS_rank'] = \
        pd.Series(np.arange(num_genes), index=gene_scores['MS'].sort_values(ascending=False).index).loc[
            gene_scores.index]

    gene_scores['oncell_size_rank'] = \
        pd.Series(np.arange(num_genes), index=gene_scores['oncell_size'].sort_values(ascending=False).index).loc[
            gene_scores.index]

    # Exclude outliers genes from ranking.
    gene_scores['voting_score_rank'][~(
            (gene_scores['log_fold_change'] > log_fold_change_thres) &
            (gene_scores['oncell_size'] > int(oncell_size_min_count)) &
            (gene_scores['oncell_size'] < int(oncell_size_max_proportion / 100 * num_cells))
    )] = len(gene_scores)

    gene_scores['bimodality_score_rank'] = gene_scores[['QQratio_rank', 'MS_rank']].min(axis=1).astype(int)
    gene_scores['bimodality_score_rank'][~(
            (gene_scores['log_fold_change'] > log_fold_change_thres) &
            (gene_scores['oncell_size'] > int(oncell_size_min_count)) &
            (gene_scores['oncell_size'] < int(oncell_size_max_proportion / 100 * num_cells))
    )] = len(gene_scores)

    gene_scores['proximity_score_rank'] = gene_scores['proximity_score_rank'].copy().astype(int)
    gene_scores['proximity_score_rank'][~(
            (gene_scores['log_fold_change'] > log_fold_change_thres) &
            (gene_scores['oncell_size'] > int(oncell_size_min_count)) &
            (gene_scores['oncell_size'] < int(oncell_size_max_proportion / 100 * num_cells))
    )] = len(gene_scores)

    MarcoPolo_score = gene_scores[['voting_score_rank', 'proximity_score_rank', 'bimodality_score_rank']].min(axis=1)

    gene_scores['MarcoPolo'] = MarcoPolo_score
    gene_scores['MarcoPolo_rank'] = pd.Series(np.arange(gene_scores.shape[0]),
                                           index=gene_scores.sort_values(['MarcoPolo', 'log_fold_change'],
                                                                      ascending=[True, False]).index).loc[
        gene_scores.index]

    gene_scores = gene_scores.reindex(sorted(gene_scores.columns), axis=1)

    return gene_scores
    
    



import numpy as np

import torch
import torch.nn as nn

torch.set_default_dtype(torch.float64)

class Masked_Function(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, mask):
        output=input
        ctx.save_for_backward(input, mask)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        input, mask = ctx.saved_tensors
        grad_input = grad_mask = None
        if ctx.needs_input_grad[0]:
            grad_input = grad_output.mul(mask)

        return grad_input, grad_mask
    
class Masked(nn.Module):    
    def __init__(self, mask):    
        super(Masked, self).__init__()
        
        self.mask = nn.Parameter(torch.Tensor(mask)==1, requires_grad=False)
        
    def forward(self, input):
        return Masked_Function.apply(input, self.mask)

    def extra_repr(self):
        return 'mask={}'.format(self.mask.shape)


class Poisson_logprob(nn.Module):
    def __init__(self):
        super(Poisson_logprob,self).__init__()
        
    def forward(self,rate,value):
        return (rate.log() * value) - rate - (value + 1).lgamma()
    
poisson_logprob=Poisson_logprob()


class Poisson_Function(torch.autograd.Function):
    @staticmethod
    def forward(ctx, Y, X, s, delta_log, beta, mask, to_return='LL'):

        with torch.no_grad():
            mu=torch.exp((X.matmul(beta)+torch.log(s.view(-1, 1))).unsqueeze(dim=1).repeat(1,delta_log.shape[0],1)+torch.exp(delta_log)*mask)
            Y_extend=Y.unsqueeze(dim=1).repeat(1,mu.shape[1],1)
            Y_logprob=poisson_logprob(rate=mu,value=Y_extend) # (N,C,G)
            Y_logprob_reduce=Y_logprob.sum(axis=2) # (N,C)
            
            Y_logprob_reduce_reduce=torch.logsumexp(Y_logprob_reduce,dim=1).view(-1,1) # (N,1)
            
            LL=torch.sum(Y_logprob_reduce_reduce) # (1)
            
            gamma=torch.exp(Y_logprob_reduce-Y_logprob_reduce_reduce)
            A=mu-Y.unsqueeze(dim=1)        
            
            #gradient
            grad_delta_log=(A*gamma.unsqueeze(dim=2)).sum(axis=0)
            grad_beta=(X.unsqueeze(dim=2)@gamma.unsqueeze(dim=1)@A).sum(axis=0)
        
            ctx.save_for_backward(grad_delta_log,grad_beta)
            
        if to_return=='LL':
            return LL
        elif to_return=='gamma':
            return gamma
        else:
            raise

    @staticmethod
    def backward(ctx, grad_output):
        
        grad_Y = grad_X = grad_s = grad_delta_log = grad_beta = grad_mask=None
        grad_delta_log,grad_beta = ctx.saved_tensors

        return grad_Y, grad_X, grad_s, grad_delta_log, grad_beta, grad_mask


class MarcoPoloModel(nn.Module):
    def __init__(self, Y, rho, X_col=5, delta_min=2):
        # Y,rho are needed for model parameter initialization
        super(MarcoPoloModel, self).__init__()

        # rho
        self.masked = Masked(rho)
        self.init_paramter_rho(rho)
        # delta
        with np.errstate(divide='ignore'):
            self.delta_log_min = np.log(delta_min)
        self.delta_log = nn.Parameter(torch.Tensor(np.ones(rho.shape)), requires_grad=True)  # (C,G)
        self.init_parameter_delta_min(delta_min)
        #beta
        self.beta=nn.Parameter(torch.Tensor(np.ones((X_col,Y.shape[1]))),requires_grad=True) # (P,G)
        self.init_paramter_Y(Y)
        
    def init_paramter_rho(self,rho):
        self.masked.mask.data=torch.Tensor((rho==1)).to(self.masked.mask.device)
        
    def init_parameter_delta_min(self,delta_min):
        with np.errstate(divide='ignore'):
            self.delta_log_min=np.log(delta_min) #
        if delta_min==0:
            self.delta_log.data=torch.Tensor(np.random.uniform(np.log(2)-0.1,np.log(2)+0.1,size=self.delta_log.shape)).to(self.delta_log.device) # (C,G)
        else:
            self.delta_log.data=torch.Tensor(np.random.uniform(self.delta_log_min-0.1,self.delta_log_min+0.1,size=self.delta_log.shape)).to(self.delta_log.device) # (C,G)
        self.delta_log.data=self.delta_log.data.clamp(min=self.delta_log_min)    

    def init_paramter_Y(self,Y):
        Y_colmean=np.mean(Y,axis=0)
        beta_init=np.hstack([((Y_colmean-Y_colmean.mean())/(np.std(Y_colmean) if len(Y_colmean)>1 else 1)).reshape(-1,1), np.zeros((Y.shape[1],self.beta.shape[0]-1))]).T
        self.beta.data[:]=torch.Tensor(beta_init).to(self.beta.device)       
        
    def forward(self, Y, X, s, to_return='LL'):
        """
        Get Y, X, and s and return LL or gamma

        Args:
            Y (torch.Tensor): Tensor of shape (number of cells, number of genes)
            X (torch.Tensor): Tensor of shape (number of cells, number of covariates)
            s (torch.Tensor): Tensor of shape (num. of cells, )
            to_return: 'LL' or 'gamma'

        Returns:
            torch.Tensor: Log-likelihood of the data
        """
        if to_return=='LL':
            delta_log_masked=self.masked(self.delta_log) #(C,G)
            LL=Poisson_Function.apply(Y, X, s, delta_log_masked, self.beta, self.masked.mask)
            return LL                      
        elif to_return=='gamma':
            with torch.no_grad():
                gamma=Poisson_Function.apply(Y, X, s, self.delta_log, self.beta, self.masked.mask, 'gamma')
            return gamma
        else:
            raise ValueError('to_return must be either "LL" or "gamma"')

if __name__ == '__main__':
    model = MarcoPoloModel(Y=np.ones((5, 5)), rho=np.ones((5, 5)))
    a = model(Y=torch.Tensor(np.ones((5, 5))), X=torch.Tensor(np.ones((5, 5))), s=torch.Tensor(np.ones((5, 1))))
    a.backward()




import os
import shutil

import numpy as np
import pandas as pd
import anndata as ad
import matplotlib.pyplot as plt
import seaborn as sns
from jinja2 import Template
from tqdm import tqdm
from typing import Union, List, Tuple
from pathlib import Path
import MarcoPolo.utils

def get_discrete_palette(n, palette=None):
    """
    Porting of https://github.com/satijalab/seurat/blob/b51801bc4b1a66aed5456473c9fe0be884994c93/R/visualization.R#L2686
    Generate a list of colors that are well separated with one another.
    
    :param int n: number of colors
    
    :return numpy.array: list of colors
    """
    palettes={
                'alphabet':[
                  "#F0A0FF", "#0075DC", "#993F00", "#4C005C", "#191919", "#005C31",
                  "#2BCE48", "#FFCC99", "#808080", "#94FFB5", "#8F7C00", "#9DCC00",
                  "#C20088", "#003380", "#FFA405", "#FFA8BB", "#426600", "#FF0010",
                  "#5EF1F2", "#00998F", "#E0FF66", "#740AFF", "#990000", "#FFFF80",
                  "#FFE100", "#FF5005"
                ],
                'alphabet2':[
                  "#AA0DFE", "#3283FE", "#85660D", "#782AB6", "#565656", "#1C8356",
                  "#16FF32", "#F7E1A0", "#E2E2E2", "#1CBE4F", "#C4451C", "#DEA0FD",
                  "#FE00FA", "#325A9B", "#FEAF16", "#F8A19F", "#90AD1C", "#F6222E",
                  "#1CFFCE", "#2ED9FF", "#B10DA1", "#C075A6", "#FC1CBF", "#B00068",
                  "#FBE426", "#FA0087"
                ],
                'glasbey':[
                  "#0000FF", "#FF0000", "#00FF00", "#000033", "#FF00B6", "#005300",
                  "#FFD300", "#009FFF", "#9A4D42", "#00FFBE", "#783FC1", "#1F9698",
                  "#FFACFD", "#B1CC71", "#F1085C", "#FE8F42", "#DD00FF", "#201A01",
                  "#720055", "#766C95", "#02AD24", "#C8FF00", "#886C00", "#FFB79F",
                  "#858567", "#A10300", "#14F9FF", "#00479E", "#DC5E93", "#93D4FF",
                  "#004CFF", "#F2F318"
                ],
                'polychrome':[
                  "#5A5156", "#E4E1E3", "#F6222E", "#FE00FA", "#16FF32", "#3283FE",
                  "#FEAF16", "#B00068", "#1CFFCE", "#90AD1C", "#2ED9FF", "#DEA0FD",
                  "#AA0DFE", "#F8A19F", "#325A9B", "#C4451C", "#1C8356", "#85660D",
                  "#B10DA1", "#FBE426", "#1CBE4F", "#FA0087", "#FC1CBF", "#F7E1A0",
                  "#C075A6", "#782AB6", "#AAF400", "#BDCDFF", "#822E1C", "#B5EFB5",
                  "#7ED7D1", "#1C7F93", "#D85FF7", "#683B79", "#66B0FF", "#3B00FB"
                ],
                'stepped':[
                  "#990F26", "#B33E52", "#CC7A88", "#E6B8BF", "#99600F", "#B3823E",
                  "#CCAA7A", "#E6D2B8", "#54990F", "#78B33E", "#A3CC7A", "#CFE6B8",
                  "#0F8299", "#3E9FB3", "#7ABECC", "#B8DEE6", "#3D0F99", "#653EB3",
                  "#967ACC", "#C7B8E6", "#333333", "#666666", "#999999", "#CCCCCC"
                ]
            }
    if palette is None:
        if n<=26:
            palette="alphabet"
        elif n<=32:
            palette="glasbey"
        else:
            palette="polychrome"
    
    palette_array= palettes[palette]
    assert n<=len(palette_array), "Not enough colours in specified palette"

    return np.array(palette_array)[np.arange(n)]


def annotate_gene_info(gene_scores, gene_query_list, gene_info, by):
    """
    Annotate gene_scores matrix with gene info.
    """

    gene_scores = gene_scores.copy()
    gene_info_select_list = []

    column_list = ['Symbol', 'description', 'Other_designations', 'type_of_gene', 'dbXrefs']

    not_matched_list = []
    pbar=tqdm(gene_query_list)
    for idx, query in enumerate(pbar):
        if by == 'ID':
            gene_info_select = gene_info[gene_info['dbXrefs'].str.contains(query, regex=False)]
        else:
            gene_info_select = gene_info[gene_info['Symbol'].str.lower() == query.lower()]
            if len(gene_info_select) == 0:
                gene_info_select = gene_info[gene_info['Synonyms'].str.lower().str.contains(query.lower(), regex=False)]

        if len(gene_info_select) >= 1:
            gene_info_select_list.append(gene_info_select[column_list].iloc[0])
        else:
            gene_info_select_list.append(pd.Series(index=column_list, dtype=float))
            not_matched_list.append(query)
        #pbar.set_description(f"Number of genes unmatched: {len(not_matched_list)}/ {len(gene_query_list)}")
        pbar.set_postfix({'Num. of unmatched genes': len(not_matched_list)})
    print(f"{len(not_matched_list)} not matched genes: {', '.join(not_matched_list[:20])+ ', ...' if len(not_matched_list) > 20 else ', '.join(not_matched_list)}")
    gene_info_extract = pd.DataFrame(gene_info_select_list, index=np.arange(len(gene_query_list)))


    assert len(gene_info_extract) == len(gene_scores), "gene_info_extract and gene_scores have different length"
    gene_scores = gene_scores.merge(right=gene_info_extract, left_index=True, right_index=True)

    return gene_scores


def generate_html_file(output_dir, gene_scores, num_genes, num_cells, top_num_html=1000):
    os.makedirs('{}'.format(output_dir), exist_ok=True)
    os.makedirs('{}/plot_image'.format(output_dir), exist_ok=True)
    os.makedirs('{}/assets'.format(output_dir), exist_ok=True)

    shutil.copy(os.path.join(os.path.dirname(__file__), 'template/assets/scripts.js'),
                '{}/assets/scripts.js'.format(output_dir))
    shutil.copy(os.path.join(os.path.dirname(__file__), 'template/assets/styles.css'),
                '{}/assets/styles.css'.format(output_dir))
    shutil.copy(os.path.join(os.path.dirname(__file__), 'template/assets/details_open.png'),
                '{}/assets/details_open.png'.format(output_dir))
    shutil.copy(os.path.join(os.path.dirname(__file__), 'template/assets/details_close.png'),
                '{}/assets/details_close.png'.format(output_dir))
    shutil.copy(os.path.join(os.path.dirname(__file__), 'template/assets/mp.png'),
                '{}/assets/mp.png'.format(output_dir))
    shutil.copy(os.path.join(os.path.dirname(__file__), 'template/assets/mp_white.png'),
                '{}/assets/mp_white.png'.format(output_dir))
    shutil.copy(os.path.join(os.path.dirname(__file__), 'template/assets/mp_white_large_font.png'),
                '{}/assets/mp_white_large_font.png'.format(output_dir))

    with open(os.path.join(os.path.dirname(__file__), 'template/index.html'), 'r') as f:
        template_read = f.read()
    template = Template(source=template_read)

    MarcoPolo_table = gene_scores.sort_values("MarcoPolo_rank", ascending=True).set_index('MarcoPolo_rank').iloc[
                      :top_num_html]
    MarcoPolo_table.index += 1
    MarcoPolo_table = MarcoPolo_table.to_html(classes="table table-bordered", table_id='dataTable')

    MarcoPolo_table = MarcoPolo_table.replace('<table ', '<table width="100%" cellspacing="0" ')
    template_rendered = template.render(MarcoPolo_table=MarcoPolo_table, num_gene=num_genes, num_cell=num_cells)

    with open('{}/index.html'.format(output_dir), 'w') as f:
        f.write(template_rendered)

def generate_image_files(adata, size_factor_key, gamma_argmax_list, gene_scores, low_dim_key, output_dir, top_num_image,
                         cell_color_key=None,
                         main_plot_s=25, main_plot_dpi=100, main_plot_font_size=10, main_plot_font_family='Arial',
                         gene_plot_s=10, gene_plot_dpi=60, gene_plot_font_size=15, gene_plot_font_family='Arial',
                         ):
    print("Drawing figures")

    plt.rcParams["figure.figsize"] = (16, 16)
    plt.rcParams["font.size"] = main_plot_font_size
    plt.rcParams['font.family'] = main_plot_font_family

    fig = plt.figure(figsize=(10, 10))
    gs = fig.add_gridspec(10, 10)

    ax = fig.add_subplot(gs[1:9, 1:9])

    if cell_color_key is None:
        plot_value = pd.Series("temp", index=adata.obs.index)
    else:
        plot_value = adata.obs[cell_color_key]
    plot_value_unique = plot_value.unique().tolist()

    expression_matrix= adata.X.copy()

    if not type(expression_matrix) == np.ndarray:
        expression_matrix = expression_matrix.toarray().astype(float)
    else:
        expression_matrix = expression_matrix.astype(float)

    if size_factor_key is not None:
        expression_matrix = (expression_matrix / adata.obs[size_factor_key].values.astype(float)[:, np.newaxis])
        print("size factor corrected")
    else:
        print("No size factor key")

    cell_meta_info=adata.obs.copy()
    cell_meta_info["coord_x"]=adata.obsm[low_dim_key][:,0]
    cell_meta_info["coord_y"] = adata.obsm[low_dim_key][:, 1]
    scatterfig = sns.scatterplot(x="coord_x", y="coord_y", hue=plot_value, data=cell_meta_info,
                                 palette=get_discrete_palette(
                                     len(plot_value_unique)).tolist() if plot_value.dtype == int else None,
                                 ax=ax, s=main_plot_s, alpha=1, edgecolor='None'
                                 )

    ax.get_legend().remove()
    ax.set_ylabel(' ')
    ax.set_xlabel(' ')

    ax.set_xticks([])
    ax.set_yticks([])

    for axis in ['top', 'bottom', 'left', 'right']:
        ax.spines[axis].set_linewidth(2)

    fig.savefig('{}/plot_image/2D_Plot.png'.format(output_dir), dpi=main_plot_dpi, bbox_inches='tight')
    plt.close(fig)


    plt.rcParams["font.size"] = gene_plot_font_size
    plt.rcParams['font.family'] = gene_plot_font_family
    plt.ioff()

    for count_idx, (iter_idx, row) in enumerate(tqdm(gene_scores.sort_values('MarcoPolo', ascending=True).iterrows(), total=top_num_image)):
        if count_idx == top_num_image:
            break

        fig = plt.figure(figsize=(3 * 8, 6))
        gs = fig.add_gridspec(6, 3 * 8)
        subplot_list = [fig.add_subplot(gs[0:6, 0:6]),
                        fig.add_subplot(gs[0:6, 6 + 2:6 + 2 + 6])]
        #import ipdb; ipdb.set_trace()
        exp_data_corrected_on = expression_matrix.T[iter_idx][gamma_argmax_list[iter_idx] == 0]
        exp_data_corrected_off = expression_matrix.T[iter_idx][gamma_argmax_list[iter_idx] == 1]

        bins_log = [np.power(1.2, i) for i in range(
            np.max([1, int(np.log(np.max([1, np.max(expression_matrix)])) / np.log(1.2))]))]

        for idx in range(2):
            ax = subplot_list[idx]
            if idx == 0:
                sns.scatterplot(x="coord_x", y="coord_y", hue=plot_value, data=cell_meta_info,
                                palette=get_discrete_palette(
                                    len(plot_value_unique)).tolist() if plot_value.dtype == int else None,
                                ax=ax, alpha=0.3, edgecolor="None",
                                s=gene_plot_s
                                )
                sns.scatterplot(x="coord_x", y="coord_y", data=cell_meta_info.loc[gamma_argmax_list[iter_idx] == 0],
                                ax=ax,
                                edgecolor=[1, 0, 0, 1],
                                facecolors="None",
                                linewidth=1,
                                s=gene_plot_s,
                                zorder=10
                                )
                ax.title.set_text('On-Off in 2D plot')
                ax.get_legend().remove()

                ax.set_xlabel('Dim 1')
                ax.set_ylabel('Dim 2', labelpad=-10)

                for axis in ['top', 'bottom', 'left', 'right']:
                    ax.spines[axis].set_linewidth(1.5)

            elif idx == 1:
                ax.title.set_text('Expression of Cells')

                ax.hist(exp_data_corrected_on, bins=bins_log, label='On', color=(1, 0, 0, 0.8))
                ax.hist(exp_data_corrected_off, bins=bins_log, label='Off', color=(0.5, 0.5, 0.5, 0.5))
                ax.set_xscale('log')

                ax.set_ylabel('Cell Counts')
                ax.set_xlabel('Expression count')

                ax.spines['right'].set_visible(False)
                ax.spines['top'].set_visible(False)

                leg = ax.legend(loc='upper left',
                                fontsize=15,
                                frameon=False,
                                bbox_to_anchor=(0.22, -0.15),
                                ncol=2,
                                handletextpad=0.2,
                                columnspacing=1.3,
                                markerscale=2.5)
                for rec in leg.get_patches():
                    rec.set_height(8)
                    rec.set_width(15)

                for axis in ['top', 'bottom', 'left', 'right']:
                    ax.spines[axis].set_linewidth(2)

        fig.savefig('{}/plot_image/{}.png'.format(output_dir, iter_idx), dpi=gene_plot_dpi, bbox_inches='tight')

        plt.close(fig)


def generate_report(adata: ad.AnnData, size_factor_key: Union[str, None], regression_result: dict, gene_scores: pd.DataFrame, output_dir: str,  low_dim_key:str, cell_color_key=None, gene_info_path: str=None,
                    top_num_html: int=1000, top_num_image: int=1000, mode=2, plot_parameters: dict={}):

    """
    Args:
        adata: anndata.AnnData containing scRNA-seq data. `.X` should be a matrix containing raw count data of shape (# cells, # genes).
        size_factor_key: key of the `adata.obs` containing the size factors. If None, no size factors will be used.
        regression_result: dict containing regression results. Return value of `run_regression` function.
        gene_scores: pd.DataFrame containing gene scores. Return value of `find_markers` function.
        gene_info_path: 'https://ftp.ncbi.nih.gov/gene/DATA/GENE_INFO/Mammalia/Homo_sapiens.gene_info.gz'
        output_dir: directory to save the report
        low_dim_key: key for accessing the 2D coordinates stored in `adata.obsm`
        cell_color_key: key for accessing the variables for coloring cells stored in `adata.obs`. If None, no cell coloring will be used.
        top_num_html: the number of top genes to show in the HTML report. Default: 1000.
        top_num_image: the number of top genes to generate figures. Default: 1000.
        mode: the number of groups used for marker selection. Default: 2.
        plot_parameters: The parameter to be passed to the `generate_image_files` function ex: {"main_plot_s": 25, "main_plot_dpi": 100, "main_plot_font_size": 10, "main_plot_font_family": 'Arial', "gene_plot_s": 10, "gene_plot_dpi": 60, "gene_plot_font_size": 15, "gene_plot_font_family": 'Arial'}

    Returns:

    """
    expression_matrix = adata.X
    num_cells=expression_matrix.shape[0]
    num_genes=expression_matrix.shape[1]

    gene_scores_munge=gene_scores.copy()
    gene_scores_munge['Gene ID']=adata.var.index.values

    output_dir=str(Path(output_dir) / "report")

    ########################
    # Assign cells to on-cells and off-cells
    ########################
    print("Assign cells to on-cells and off-cells...")
    gamma_list = regression_result["gamma_list_cluster"][mode]
    gamma_argmax_list = MarcoPolo.utils.gamma_list_expression_matrix_to_gamma_argmax_list(gamma_list, expression_matrix)

    ########################
    # Calculate voting score
    ########################
    #oncell_size_list = np.sum(gamma_argmax_list == 0, axis=1)
    oncell_size_cliplist = MarcoPolo.utils.gamma_argmax_list_to_oncell_size_list_list(gamma_argmax_list)
    intersection_list = MarcoPolo.utils.gamma_argmax_list_to_intersection_list(gamma_argmax_list)
    intersectioncount_prop=((intersection_list/oncell_size_cliplist))
    intersectioncount_prop_top10=[np.arange(0,len(i))[i>=sorted(i)[-10]][:10] for i in intersectioncount_prop]
    
    gene_scores_munge['Voting_genes_top10']=[gene_scores_munge['Gene ID'][i].values for i in intersectioncount_prop_top10]
    gene_scores_munge_voting=gene_scores_munge.copy()    

    ########################
    # Annotate gene_scores with gene info
    ########################
    if gene_info_path is not None:
        print(f"Annotating genes with the gene info: {gene_info_path}")
        
        gene_info=pd.read_csv(gene_info_path,sep='\t')
        #import ipdb
        #ipdb.set_trace()
        by='ID' if 'ENS' in adata.var.index.values.tolist()[0] else 'name'
        gene_scores_munge=annotate_gene_info(gene_scores=gene_scores_munge, gene_query_list=adata.var.index.values.tolist(), gene_info=gene_info, by=by)

        gene_scores_munge['Log2FC']=(gene_scores_munge['log_fold_change']/np.log10(2)).round(2)

        gene_scores_munge=gene_scores_munge[[
                                    'MarcoPolo_rank',
                                    'Gene ID',
                                    'Symbol', 'description', 'Other_designations', 'type_of_gene',
                                    'Log2FC',
                                    'MarcoPolo',
                                    'bimodality_score_rank',
                                    'voting_score_rank',
                                    'proximity_score_rank',
                                    'oncell_size','oncell_size_rank',
                                    'dbXrefs'
                                   ]]
        gene_scores_munge['img'] = gene_scores_munge.apply(lambda x: '<img src="plot_image/{idx}.png" alt="{idx}">'.format(idx=x.name), axis=1)
        import ipdb; ipdb.set_trace()

    else:
        gene_scores_munge['Log2FC'] = (gene_scores_munge['log_fold_change'] / np.log10(2)).round(2)

        gene_scores_munge=gene_scores_munge[[
                                    'MarcoPolo_rank',
                                    'Gene ID',
                                    'Log2FC',
                                    'MarcoPolo',
                                    'bimodality_score_rank',
                                    'voting_score_rank',
                                    'proximity_score_rank',
                                    'oncell_size','oncell_size_rank',
                                   ]]
        gene_scores_munge['img']=gene_scores_munge.apply(lambda x: '<img src="plot_image/{idx}.png" alt="{idx}">'.format(idx=x.name),axis=1)


    ########################
    # Generate table files
    ########################
    print(f"Generating table files...")
    generate_html_file(output_dir=output_dir, gene_scores=gene_scores_munge, num_genes=num_genes, num_cells=num_cells, top_num_html=top_num_html)
    gene_scores_munge_voting[['Gene ID','Voting_genes_top10']].to_html('{}/voting_result.html'.format(output_dir))
    gene_scores_munge.to_csv('{}.table.tsv'.format(output_dir), sep='\t')

    ########################
    # Generate image files
    ########################
    print(f"Generating image files...")
    generate_image_files(adata=adata, size_factor_key=size_factor_key, gamma_argmax_list=gamma_argmax_list, gene_scores=gene_scores,
                         low_dim_key=low_dim_key, output_dir=output_dir, top_num_image=top_num_image,
                         cell_color_key=cell_color_key,
                         **plot_parameters)











from setuptools import setup, find_packages

# python -m build
# python -m twine upload --repository pypi dist/* --verbose

setup(
  name = 'marcopolo-pytorch',
  packages = find_packages(exclude=[]),
  include_package_data = True,
  version = '1.0.8',
  description = 'MarcoPolo - Pytorch',
  author = 'Chanwoo Kim',
  author_email = 'ch6845@gmail.com',
  long_description_content_type = 'text/markdown',
  url = 'https://github.com/chanwkimlab/MarcoPolo',
  keywords = [
    'single-cell',
    'bioinformatics',
    'pytorch'
  ],
  install_requires=[
    'tqdm',
    'einops>=0.3',
    'numpy>=1.19.2',
    'torch>=1.4.0',
    'pandas>=1.2.0',
    'scikit-learn>=0.24.1',
    'scipy>=1.6.1',
    'matplotlib>=3.3.0',
    'seaborn>=0.11.1',
    'Jinja2>=2.11.2',
    'anndata>=0.7.4',
    'rpy2>=3.4.2',
    'ipywidgets>=7.5.1',
    'scanpy>=1.9.0',
  ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'License :: Other/Proprietary License',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
  ],
)